import 'dart:async';
import 'dart:convert';

import 'package:verival/update.dart';
import 'package:verival/verival.dart' as verival;
import 'package:webroker/send.dart' as webroker;
import 'package:webroker/target.dart';
import 'package:yona_crypto/b64.dart';
import 'package:yona_crypto/crypto.dart';
import 'package:yona_messenger/peer.dart';

import 'msg.dart';

class Node {
  final B64 sk;

  final Peer peer;
  final Map<double, Completer<void>> _awaiters = {};

  Future<void> onMsg(Msg msg) async {
    print(msg);
  }

  Future<Target> lookup(Peer peer) async {
    var update =
        await verival.get(peer.verival.host, peer.verival.port, peer.pk);
    var webroker = Target.fromString(update.data["webroker"]!);
    peer.webroker = webroker;
    return webroker;
  }

  Future<void> publish() async {
    var update = Update.create(peer.pk, DateTime.now().toUtc(), {
      "webroker": peer.webroker!.toString(),
    });
    await verival.put(
        peer.verival.host, peer.verival.port, await update.encode(sk));
  }

  Node(this.sk, this.peer) : assert(peer.pk == derivePK(sk));

  Future<void> input(List<int> data) async {
    var input = Uri.splitQueryString(utf8.decode(data));
    var from = Peer.fromString(input["from"]!);
    var to = Peer.fromString(input["to"]!);
    assert(to == peer);
    var key = await deriveKey(sk, from.pk);
    var payload =
        Uri.splitQueryString(utf8.decode(await decrypt(key, input["cipher"]!)));
    switch (payload["kind"]) {
      case "data":
        var msg = Msg(
          from,
          utf8.encode(payload["data"]!),
          DateTime.fromMicrosecondsSinceEpoch(
              (double.parse(payload["time"]!) * 1000000).floor()),
        );
        if (payload["ack"]?.toLowerCase() == "true") {
          var ack = Uri(queryParameters: {
            "kind": "ack",
            "msgTime": payload["time"]!,
          }).query;
          var key = await deriveKey(sk, from.pk);
          var cipher = await encrypt(key, utf8.encode(ack));
          var output = Uri(queryParameters: {
            "from": peer.toString(),
            "to": to.toString(),
            "cipher": cipher,
          }).query;
          await webroker.send(from.webroker!, utf8.encoder.convert(output));
        }
        await onMsg(msg);
        break;
      case "ack":
        var msgId = double.parse(payload["msgTime"]!);
        _awaiters[msgId]!.complete();
        break;
      default:
        throw UnsupportedError("unsupported data kind");
    }
  }

  Future<void> send(Peer to, List<int> data, [bool ack = true]) async {
    var msgId = DateTime.now().toUtc().microsecondsSinceEpoch / 1000000;
    Timer? timer;
    try {
      to.webroker = await lookup(to);
      var payload = Uri(queryParameters: {
        "kind": "data",
        "ack": ack.toString(),
        "data": utf8.decode(data),
        "time": msgId.toString()
      }).query;
      var key = await deriveKey(sk, to.pk);
      var cipher = await encrypt(key, utf8.encode(payload));
      var output = Uri(queryParameters: {
        "from": peer.toString(),
        "to": to.toString(),
        "cipher": cipher,
      }).query;
      await webroker.send(to.webroker!, utf8.encoder.convert(output));
      var completer = Completer<void>();
      timer = Timer(Duration(seconds: 5), () {
        completer.completeError(TimeoutException("ack timed out"));
      });
      _awaiters[msgId] = completer;
      await completer.future;
    } finally {
      _awaiters.remove(msgId);
      timer?.cancel();
    }
  }
}
